/* ptiff.c */
#define PTIFF

#include <sys/types.h>
#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <tiffio.h>
#include "print.h"

#ifdef __MINGW32__
#include <windows.h>
#endif

static char *fname;
static int printModeType;
static TIFF *out;
static int tiffLZW = FALSE; /* chooes LZW or DEFLATE compression */

static int initPrint(int printType, int complexImage, int xpixels,
	int ypixels, int xdpi, int ydpi, int xmargin, int ymargin)
{
 int samplesPP=1, bitsPP=1;
 unsigned short photometric=0;
 int filedes = -1;

 printModeType = printType;

 out = NULL;
 switch (printModeType)
  {
   case IMG_BW:
    photometric = PHOTOMETRIC_MINISBLACK;
    samplesPP = 1;
    bitsPP = 1;
    break;
   case IMG_CMAP:
    photometric = PHOTOMETRIC_PALETTE;
    samplesPP = 1;
    bitsPP = 8;
    break;
   case IMG_TRUE_COL:
    photometric = PHOTOMETRIC_RGB;
    samplesPP = 3;
    bitsPP = 8;
    break;
   case IMG_RGBA:
    photometric = PHOTOMETRIC_RGB;
    samplesPP = 4;
    bitsPP = 8;
    break;
   case IMG_LAB:
    photometric = PHOTOMETRIC_CIELAB;
    samplesPP = 3;
    bitsPP = 8;
    break;
  }

#ifdef __MINGW32__
 char tempPath[MAX_PATH + 1];

 if (GetTempPathA(MAX_PATH + 1, tempPath) == 0)
  {
   // Failed to get temp path.
   perror("plib: failed to get temp path");
   return (PRT_IO_ERROR);
  }

 // This is allocated in advance because of the code below.
 char tempFilename[MAX_PATH + 1];

 fname = tempFilename;
 if (GetTempFileNameA(tempPath, "prt", 1, fname) == 0)
  {
   perror("plib: failed to create tmpfile");
   return(PRT_IO_ERROR);
  }
 fname = strdup(fname);
#else
 fname = strdup("/tmp/prtXXXXXX");
 if ((filedes=mkstemp(fname)) < 0 )
  return(PRT_IO_ERROR);
 else
  close(filedes); /* will open the file with tiff function below */
#endif  

 if ((out = TIFFOpen(fname, "wb")) == NULL)
  return(PRT_IO_ERROR);

 TIFFSetField(out, TIFFTAG_IMAGEWIDTH, (unsigned int) xpixels);
 TIFFSetField(out, TIFFTAG_IMAGELENGTH, (unsigned int) ypixels);
 TIFFSetField(out, TIFFTAG_ORIENTATION, ORIENTATION_TOPLEFT);
 TIFFSetField(out, TIFFTAG_SAMPLESPERPIXEL, samplesPP);
 TIFFSetField(out, TIFFTAG_BITSPERSAMPLE, bitsPP);
 TIFFSetField(out, TIFFTAG_PLANARCONFIG, PLANARCONFIG_CONTIG);
 TIFFSetField(out, TIFFTAG_PHOTOMETRIC, photometric);
 if (printModeType==IMG_RGBA)
  {
   uint16 extra[1];

   extra[0] = EXTRASAMPLE_ASSOCALPHA;
   TIFFSetField(out, TIFFTAG_EXTRASAMPLES, 1, &extra);
  }
 TIFFSetField(out, TIFFTAG_COMPRESSION, tiffLZW ? COMPRESSION_LZW : COMPRESSION_DEFLATE); /* was COMPRESSION_DEFLATE due to patents, but now it is expred so change it back to LZW */
 TIFFSetField(out, TIFFTAG_ROWSPERSTRIP, 1);
 TIFFSetField(out, TIFFTAG_SOFTWARE, "Arahne");
 TIFFSetField(out, TIFFTAG_RESOLUTIONUNIT, RESUNIT_INCH);
 TIFFSetField(out, TIFFTAG_XRESOLUTION, xdpi/1000.0);
 TIFFSetField(out, TIFFTAG_YRESOLUTION, ydpi/1000.0);
 return(PRT_OK);
}

static int printBand(int xpixels, char *data, int bpl, int ypos,
 int lines, char bandType, unsigned char *cmap)
{
 int x, y;

 if (data==NULL)
  return(PRT_INTERNAL_ERROR);
 if (printModeType<bandType)
  return(PRT_WRONG_IMG);

 if ((ypos==0)&&(printModeType == IMG_CMAP))
  {
   unsigned short *r, *g, *b;
   int i;

   r = (unsigned short *) malloc(sizeof(unsigned short)*3*256);
   g = r + 256;
   b = g + 256;
   for (i = 0; i < 256; i++)
    {
     r[i] = (unsigned short) (cmap[i*3]*256);
     g[i] = (unsigned short) (cmap[i*3+1]*256);
     b[i] = (unsigned short) (cmap[i*3+2]*256);
    }
   TIFFSetField(out, TIFFTAG_COLORMAP, r, g, b);
   free(r);
  }

 if ((bandType==IMG_BW)&&(printModeType==IMG_BW))
  for (y=lines*bpl-1; y>=0; y--)
   data[y] = (unsigned char) (~data[y]);
 
 switch (printModeType)
  {
   case IMG_BW:
    for (y=0; y<lines; y++)
      TIFFWriteScanline(out, &data[y*bpl], y+ypos, 0);
    break;
   case IMG_CMAP:
    if (bandType==IMG_BW)
     {
      unsigned char *dat, cdata[bpl*8];

      for (y=0; y<lines; y++)
       {
        dat = (unsigned char *) &data[y*bpl];
        for (x=0; x<bpl*8; x++)
	 cdata[x] = (unsigned char) ((dat[x>>3]&(128>>(x&7))) ? 255 : 254);
        TIFFWriteScanline(out,  cdata, y+ypos, 0);
       }
     }
    else /* (bandType==IMG_CMAP) */
    for (y=0; y<lines; y++)
     TIFFWriteScanline(out,  &data[y*bpl], y+ypos, 0);
    break;
   case IMG_TRUE_COL:
   case IMG_LAB:
    switch (bandType)
     {
      case IMG_BW:
       {
       	unsigned char *dat, *cdat, cdata[bpl*8*3];

	for (y=0; y<lines; y++)
	 {
	  cdat = cdata;
          dat = (unsigned char *) &data[y*bpl];
          if (printModeType==IMG_TRUE_COL)
	   {
            for (x=0; x<bpl*8; x++)
	     if (!(dat[x>>3]&(128>>(x&7))))
	      { *cdat++ = 255; *cdat++ = 255; *cdat++ = 255; }
	     else
	      { *cdat++ = 0; *cdat++ = 0; *cdat++ = 0; }
	   }
	  else
	   {
            for (x=0; x<bpl*8; x++)
	     if (!(dat[x>>3]&(128>>(x&7))))
	      { *cdat++ = 255; *cdat++ = 0; *cdat++ = 0; }
	     else
	      { *cdat++ = 0; *cdat++ = 0; *cdat++ = 0; }
	   }
	    
          TIFFWriteScanline(out,  cdata, y+ypos, 0);
	 }
       }
       break;
      case IMG_CMAP:
       {
       	unsigned char *dat, *cdat, cdata[bpl*8*3];

	for (y=0; y<lines; y++)
	 {
	  cdat = cdata;
          dat = (unsigned char *) &data[y*bpl];
          for (x=0; x<bpl; x++)
	   { *cdat++ = cmap[*dat*3]; *cdat++ = cmap[*dat*3+1]; *cdat++ = cmap[*dat++ *3+2]; }
          TIFFWriteScanline(out,  cdata, y+ypos, 0);
	 }
       }
       break;
      case IMG_TRUE_COL_DITH:
      case IMG_TRUE_COL:
      case IMG_LAB:
       for (y=0; y<lines; y++)
	TIFFWriteScanline(out, &data[y*bpl], y+ypos, 0);
       break;
     }
    break;

   case IMG_RGBA:
    switch (bandType)
     {
      case IMG_BW:
       {
       	unsigned char *dat, *cdat, cdata[xpixels*4];

	for (y=0; y<lines; y++)
	 {
	  cdat = cdata;
          dat = (unsigned char *) &data[y*bpl];
          for (x=0; x<xpixels; x++)
	   if (!(dat[x>>3]&(128>>(x&7))))
	    { *cdat++ = 255; *cdat++ = 255; *cdat++ = 255; *cdat++ = 255;  }
	   else
	    { *cdat++ = 0; *cdat++ = 0; *cdat++ = 0; *cdat++ = 255; }
	  TIFFWriteScanline(out, cdata, y+ypos, 0);
	 }
       }
       break;
      case IMG_CMAP:
       {
       	unsigned char *dat, *cdat, cdata[xpixels*4];

	for (y=0; y<lines; y++)
	 {
	  cdat = cdata;
          dat = (unsigned char *) &data[y*bpl];
          for (x=xpixels; x; x--)
	   { *cdat++ = cmap[*dat*3]; *cdat++ = cmap[*dat*3+1]; *cdat++ = cmap[*dat++ *3+2]; *cdat++ = 255; }
	  TIFFWriteScanline(out, cdata, y+ypos, 0);
	 }
       }
       break;
      case IMG_TRUE_COL_DITH:
      case IMG_TRUE_COL:
       {
       	unsigned char *dat, *cdat, cdata[xpixels*4];

	for (y=0; y<lines; y++)
	 {
	  cdat = cdata;
          dat = (unsigned char *) &data[y*bpl];
          for (x=xpixels; x; x--)
	   { *cdat++ = *dat++; *cdat++ = *dat++; *cdat++ = *dat++; *cdat++ = 255; }
	  TIFFWriteScanline(out, cdata, y+ypos, 0);
	 }
       }
       break;
      case IMG_RGBA:
       for (y=0; y<lines; y++)
	TIFFWriteScanline(out, &data[y*bpl], y+ypos, 0);
       break;
     }
    break;
  } 
 return(PRT_OK);
}

static int endPrint(int printStatus)
{
 (void) TIFFClose(out);
 if (printStatus==PRT_OK)
  printStatus = enqueuePrintFile(fname);

 /* just cleanup; called even if printing fails */
 removeFileFreeFilename(fname, printStatus);
 return(printStatus);
}

/* fill in the preallocated printer modes */
static void getPrtInfo(printerMode *printerMod, int *numModes,
	printerDensity *prtDensityX, int *numDensitX,
	printerDensity *prtDensityY, int *numDensitY,
	int *nonprintableTop, int *nonprintableBottom, int *nonprintableLeft,
	int *dpiVariable, char **prtName, char **profileName)
{
 int i;

 printerMod[0].printType = IMG_TRUE_COL;
 printerMod[1].printType = IMG_CMAP;
 printerMod[2].printType = IMG_BW;
 printerMod[3].printType = IMG_RGBA;
 printerMod[4].printType = IMG_LAB;

 *numModes=5;

 *numDensitX = *numDensitY = 7;
 for (i=0; i<6; i++)
  {
   prtDensityX[i].pixels = prtDensityY[i].pixels = MAX_PRINT;
   prtDensityX[i].dpi = prtDensityY[i].dpi = 100000*(i+1);
  }
 prtDensityX[6].pixels = prtDensityY[6].pixels = MAX_PRINT;
 prtDensityX[6].dpi = prtDensityY[6].dpi = 720000;

 *nonprintableTop=0;
 *nonprintableBottom=0;
 *nonprintableLeft=0;
 *dpiVariable=FALSE;
 *prtName="TIFF";
 *profileName=NULL;
}

/* register TIFF "printer" */
void registerTIFF()
{
 registerPrinter(*getPrtInfo, *initPrint, *printBand, *endPrint);
}

/* choose between default LZW or DEFLATE; if you don't set it, we use DEFLATE */
DPTIFF void setTiffLZW(int set)
{
 tiffLZW = set;
}

